# cvt/metrics.py

"""Module including routines computing metrics.

This module includes the following functions:

- `abs_error(est_depth, gt_depth)` - Computes the absolute error between an estimated and groun-truth depth map.
- `accuracy_eval(est_ply, gt_ply, mask_th, est_filt=None, gt_filt=None)` - Computes the accuracy of an estimated point cloud against the provided ground-truth.
- `completeness_eval(est_ply, gt_ply, mask_th=20.0, est_filt=None, gt_filt=None)` - Computes the completeness of an estimated point cloud against the provided ground-truth.
- `filter_outlier_points(est_ply, gt_ply, outlier_th)` - Filters out points from an estimated point cloud that are farther than some threshold to the ground-truth point cloud.
"""

import numpy as np
import open3d as o3d
from sklearn.neighbors import KDTree
from typing import Tuple, Optional
import torch

def MAE(estimate, target, reduction_dims, mask=None, relative=False):
    assert(estimate.shape==target.shape)
    error = estimate - target
    if relative:
        error /= target
    error = torch.abs(error)

    if mask != None:
        assert(error.shape==mask.shape)
        error *= mask
        error = (error.sum(dim=reduction_dims) / (mask.sum(dim=reduction_dims)+1e-10)).sum()
    else:
        error =  error.mean()
    return error

def RMSE(estimate, target, mask=None, relative=False):
    assert(estimate.shape==target.shape)
    error = estimate - target
    if relative:
        error /= target
    error = torch.square(error)

    assert(error.shape==mask.shape)
    error *= mask

    reduction_dims = tuple(range(1,len(target.shape)))
    error = (error.sum(dim=reduction_dims) / (mask.sum(dim=reduction_dims)+1e-10)).sum()
    return torch.sqrt(error)

def abs_error(est_depth: np.ndarray, gt_depth: np.ndarray) -> np.ndarray:
    """Computes the absolute error between an estimated and groun-truth depth map.

    Parameters:
        est_depth: Estimated depth map.
        gt_depth: Ground-truth depth map.

    Returns:
       The absolute error map for the estimated depth map. 
    """
    signed_error = est_depth - gt_depth

    # compute gt mask and number of valid pixels
    gt_mask = np.not_equal(gt_depth, 0.0).astype(np.double)
    error = np.abs(signed_error) * gt_mask

    return error

def accuracy_eval(
        est_ply: np.ndarray,
        gt_ply: np.ndarray,
        mask_th: float,
        est_filt: Optional[np.ndarray] = None,
        gt_filt: Optional[np.ndarray] = None) -> Tuple[o3d.geometry.PointCloud, np.ndarray, np.ndarray]:
    """Computes the accuracy of an estimated point cloud against the provided ground-truth.

    Parameters:
        est_ply: Estimated point cloud to be evaluated.
        gt_ply: Ground-truth point cloud.
        mask_th: Masking threshold used to remove points from the evaluation farther 
                    than a specified distance value.
        est_filt: Optional filter to remove unwanted point from the estimated 
                    point cloud in the evaluation
        gt_filt: Optional filter to remove unwanted point from the ground-truth
                    point cloud in the evaluation

    Returns:
        valid_est_ply: Point cloud containing all the valid evaluation points after filtering.
        dists_est: Estimated distances of all valid points in the estimated point cloud 
                    to the closest point in the ground-truth point cloud.
        colors_est: Estimated colors for the points in the valid point cloud.
    """
    # distance from est to gt
    dists_est = np.asarray(est_ply.compute_point_cloud_distance(gt_ply))

    # extract valid indices
    valid_inds_est = set(np.where(est_filt == 1)[0])
    valid_dists = set(np.where(dists_est <= mask_th)[0])
    valid_inds_est.intersection_update(valid_dists)
    valid_inds_est = np.asarray(list(valid_inds_est))

    # get distances and colors at valid indices
    valid_est_ply = est_ply.select_by_index(valid_inds_est)
    dists_est = dists_est[valid_inds_est]
    colors_est = np.asarray(est_ply.colors)[valid_inds_est]

    return valid_est_ply, dists_est, colors_est


def completeness_eval(
        est_ply: o3d.geometry.PointCloud,
        gt_ply: o3d.geometry.PointCloud,
        mask_th: float = 20.0,
        est_filt: Optional[np.ndarray] = None,
        gt_filt: Optional[np.ndarray] = None) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Computes the completeness of an estimated point cloud against the provided ground-truth.

    Parameters:
        est_ply: Estimated point cloud to be evaluated.
        gt_ply: Ground-truth point cloud.
        mask_th: Masking threshold used to remove points from the evaluation farther 
                    than a specified distance value.
        est_filt: Optional filter to remove unwanted point from the estimated 
                    point cloud in the evaluation
        gt_filt: Optional filter to remove unwanted point from the ground-truth
                    point cloud in the evaluation

    Returns:
        ply_points: Point cloud vertices containing all the valid evaluation points after filtering.
        dists: Distances of all valid points in the ground-truth point cloud 
                    to the closest point in the estimated point cloud.
        colors: Colors for the points in the valid point cloud.
    """
    # build KD-Tree of estimated point cloud for querying
    tree = KDTree(np.asarray(est_ply.points), leaf_size=40)
    (dists, inds) = tree.query(np.asarray(gt_ply.points), k=1)

    # extract valid indices
    valid_inds = set(np.where(gt_filt == 1)[0])
    valid_inds.intersection_update(set(np.where(dists <= mask_th)[0]))
    valid_inds = np.asarray(list(valid_inds))

    dists = dists[valid_inds]
    inds = inds[valid_inds]
    ply_points = np.asarray(est_ply.points)[inds]
    colors = np.asarray(est_ply.colors)[inds]

    return ply_points, dists, colors


def filter_outlier_points(est_ply: o3d.geometry.PointCloud, gt_ply: o3d.geometry.PointCloud, outlier_th: float) -> o3d.geometry.PointCloud:
    """Filters out points from an estimated point cloud that are farther than some threshold to the ground-truth point cloud.

    Parameters:
        est_ply: Estimated point cloud to filter.
        gt_ply: Ground-truth point cloud for reference.
        outlier_th: Distance threshold used for filtering.

    Returns:
        The filtered point cloud.
    """
    dists_est = np.asarray(est_ply.compute_point_cloud_distance(gt_ply))
    valid_dists = np.where(dists_est <= outlier_th)[0]
    return est_ply.select_by_index(valid_dists)



#   def roc_curve(in_depth, out_depth, in_conf, out_conf, gt_depth, path, stats_file, view_num, scene, device):
#       if (scene == "Barn"):
#           th = 0.01
#       elif(scene == "Caterpillar"):
#           th = 0.005
#       elif(scene == "Church"):
#           th = 0.025
#       elif(scene == "Courthouse"):
#           th = 0.025
#       elif (scene =="Ignatius"):
#           th = 0.003
#       elif(scene == "Meetingroom"):
#           th = 0.01
#       elif(scene == "Truck"):
#           th = 0.005
#       elif(scene == "none"): # DTU
#           th = 2
#       else:
#           th = 0.02
#   
#       height, width = in_depth.shape
#   
#       valid_map = torch.ne(gt_depth, 0.0).to(torch.float32).to(device)
#       valid_count = torch.sum(valid_map)+1e-7
#   
#       # flatten to 1D tensor
#       in_depth = torch.flatten(in_depth)
#       in_conf = torch.flatten(in_conf)
#       out_depth = torch.flatten(out_depth)
#       out_conf = torch.flatten(out_conf)
#       gt_depth = torch.flatten(gt_depth)
#   
#       ##### INPUT #####
#       # sort all tensors by confidence value
#       (in_conf,indices) = in_conf.sort(descending=True)
#       in_depth = torch.gather(in_depth, dim=0, index=indices)
#       in_gt_depth = torch.gather(gt_depth, dim=0, index=indices)
#       # pull only gt values
#       in_gt_indices = torch.nonzero(in_gt_depth).flatten()
#       in_depth = torch.index_select(in_depth, dim=0, index=in_gt_indices)
#       in_gt_depth = torch.index_select(in_gt_depth, dim=0, index=in_gt_indices)
#   
#       # sort orcale curves by error
#       in_oracle = torch.abs(in_depth-in_gt_depth)
#       (in_oracle,indices) = in_oracle.sort(descending=False)
#       in_oracle_gt = torch.gather(in_gt_depth, dim=0, index=indices)
#       # pull only gt values
#       in_oracle_indices = torch.nonzero(in_oracle_gt).flatten()
#       in_oracle = torch.index_select(in_oracle, dim=0, index=in_oracle_indices)
#   
#       ##### OUTPUT #####
#       # sort all tensors by confidence value
#       (out_conf,indices) = out_conf.sort(descending=True)
#       out_depth = torch.gather(out_depth, dim=0, index=indices)
#       out_gt_depth = torch.gather(gt_depth, dim=0, index=indices)
#       # pull only gt values
#       out_gt_indices = torch.nonzero(out_gt_depth).flatten()
#       out_depth = torch.index_select(out_depth, dim=0, index=out_gt_indices)
#       out_gt_depth = torch.index_select(out_gt_depth, dim=0, index=out_gt_indices)
#   
#       out_oracle = torch.abs(out_depth-out_gt_depth)
#       (out_oracle,indices) = out_oracle.sort(descending=False)
#       out_oracle_gt = torch.gather(out_gt_depth, dim=0, index=indices)
#       # pull only gt values
#       out_oracle_indices = torch.nonzero(out_oracle_gt).flatten()
#       out_oracle = torch.index_select(out_oracle, dim=0, index=out_oracle_indices)
#   
#       # build density vector
#       num_gt_points = in_gt_depth.shape[0]
#       perc = np.array(list(range(5,105,5)))
#       density = np.array((perc/100) * (num_gt_points), dtype=np.int32)
#   
#       in_prec = np.zeros(density.shape)
#       in_prec_oracle = np.zeros(density.shape)
#       out_prec = np.zeros(density.shape)
#       out_prec_oracle = np.zeros(density.shape)
#   
#       in_rec = np.zeros(density.shape)
#       in_rec_oracle = np.zeros(density.shape)
#       out_rec = np.zeros(density.shape)
#       out_rec_oracle = np.zeros(density.shape)
#   
#       for i,k in enumerate(density):
#           # compute input absolute error chunk
#           iae = torch.abs(in_gt_depth[0:k] - in_depth[0:k])
#           num_inliers = torch.sum(torch.le(iae, th).to(torch.float32).to(device))
#   
#           in_prec[i] = num_inliers / k
#           in_rec[i] = num_inliers / valid_count
#   
#           # compute input oracle chunk
#           in_prec_oracle[i] = torch.sum(torch.le(in_oracle[0:k], th).to(torch.float32).to(device)) / k
#           in_rec_oracle[i] = torch.sum(torch.le(in_oracle[0:k], th).to(torch.float32).to(device)) / valid_count
#   
#           # compute output absolute error chunk
#           oae = torch.abs(out_gt_depth[0:k] - out_depth[0:k])
#           num_inliers = torch.sum(torch.le(oae, th).to(torch.float32).to(device))
#           out_prec[i] = num_inliers / k
#           out_rec[i] = num_inliers / valid_count
#   
#           # compute output oracle chunk
#           out_prec_oracle[i] = torch.sum(torch.le(out_oracle[0:k], th).to(torch.float32).to(device)) / k
#           out_rec_oracle[i] = torch.sum(torch.le(out_oracle[0:k], th).to(torch.float32).to(device)) / valid_count
#   
#       in_prec_str = ",".join([ "{:0.5f}".format(r) for r in in_prec ])
#       out_prec_str = ",".join([ "{:0.5f}".format(r) for r in out_prec ])
#       in_prec_oracle_str = ",".join([ "{:0.5f}".format(r) for r in in_prec_oracle ])
#       out_prec_oracle_str = ",".join([ "{:0.5f}".format(r) for r in out_prec_oracle ])
#   
#       in_rec_str = ",".join([ "{:0.5f}".format(r) for r in in_rec ])
#       out_rec_str = ",".join([ "{:0.5f}".format(r) for r in out_rec ])
#       in_rec_oracle_str = ",".join([ "{:0.5f}".format(r) for r in in_rec_oracle ])
#       out_rec_oracle_str = ",".join([ "{:0.5f}".format(r) for r in out_rec_oracle ])
#   
#       # store data
#       with open(stats_file,'a') as f:
#           f.write("{}\n".format(in_prec_str))
#           f.write("{}\n".format(out_prec_str))
#           f.write("{}\n".format(in_prec_oracle_str))
#           f.write("{}\n".format(out_prec_oracle_str))
#   
#           f.write("{}\n".format(in_rec_str))
#           f.write("{}\n".format(out_rec_str))
#           f.write("{}\n".format(in_rec_oracle_str))
#           f.write("{}\n".format(out_rec_oracle_str))
#   
#       return



# cvt/camera.py

"""A suite of common camera utilities.

This module includes several functions for manipulating and extracting information
from camera intrinsics and extrinsics, as well as converting between specific
formats.

This module contains the following functions:

- `camera_center(cam)` - Computes the center of a camera in world coordinates.
- `intrinsic_pyramid(K, levels)` - Scales camera intrinsics for multi-resolution operations.
- `relative_transform(cams_1, cams_2)` - Computes the relative transformation between two sets of cameras.
- `sfm_to_trajectory(cams, log_file)` - Convert a set of cameras from SFM format to Trajectory File format.
- `trajectory_to_sfm(log_file, camera_path, intrinsics)` - Convert a set of cameras from Trajectory File format to SFM format.
- `y_axis_rotation(P, theta)` - Applies a rotation to the given camera extrinsics matrix along the y-axis.
"""

import os
import numpy as np
import math
from scipy.linalg import null_space
from scipy.spatial.transform import Rotation
import torch
import open3d as o3d

def crop_cam(intrinsics, crop_row, crop_col):
    intrinsics[0,2] -= crop_col
    intrinsics[1,2] -= crop_row
    return intrinsics


def scale_cam(intrinsics, h=None, w=None, max_h=None, max_w=None, scale=None):
    if scale:
        new_intrinsics = intrinsics.copy()
        new_intrinsics[0, :] *= scale
        new_intrinsics[1, :] *= scale
    elif h > max_h or w > max_w:
        scale = 1.0 * max_h / h
        if scale * w > max_w:
            scale = 1.0 * max_w / w
        new_intrinsics = intrinsics.copy()
        new_intrinsics[0, :] *= scale
        new_intrinsics[1, :] *= scale
    return new_intrinsics

def build_o3d_traj(poses, K, width, height):
    trajectory = o3d.camera.PinholeCameraTrajectory()
    for pose in poses:
        camera_params = o3d.camera.PinholeCameraParameters()
        camera_params.intrinsic = o3d.camera.PinholeCameraIntrinsic(width, height, K[0,0], K[1,1], K[0,2], K[1,2])
        camera_params.extrinsic = pose
        trajectory.parameters += [(camera_params)]
    return trajectory

def to_opengl_pose(pose):
    """
    OpenGL pose: (right-up-back) (cam-to-world)
    """
    #pose = torch.linalg.inv(pose)
    pose = np.linalg.inv(pose)
    pose[:3,1] *= -1
    pose[:3,2] *= -1
    return pose

def intrinsic_pyramid(K, levels):
    batch_size, _, _ = K.shape
    K_pyr = torch.zeros((batch_size, levels, 3, 3)).to(K)

    for l in range(levels):
        if l==0:
            k = K
        else:
            k = torch.clone(K) / (2**l)
            k[:,2,2] = 1.0
        K_pyr[:,l] = k

    return K_pyr

def scale_intrinsic(K, s):
    batch_size, _, _ = K.shape

    Ks = torch.clone(K) / (2**(s-1))
    Ks[:,2,2] = 1.0

    return Ks

def _intrinsic_pyramid(K, levels):
    K_pyr = np.zeros((levels, 3, 3))

    for l in range(levels):
        if l==0:
            k = K
        else:
            k = np.copy(K) / (2**l)
            k[2,2] = 1.0
        K_pyr[l] = k

    return K_pyr

def Z_from_disp(Z, b, f, delta):
    B, C, D, H, W = Z.shape
    b = b.reshape(B,1,1,1,1).repeat(1,C,D,H,W)
    f = f.reshape(B,1,1,1,1).repeat(1,C,D,H,W)

    near = Z*(b*f/((b*f) + (delta*Z)))
    far = Z*(b*f/((b*f) - (delta*Z)))

    return near, far

def compute_baselines(poses):
    src_pose = poses[0]
    src_camera_center = null_space(src_pose[:3,:])
    src_camera_center = src_camera_center[:3,0] / src_camera_center[3,0]
    min_baseline = np.Inf
    max_baseline = 0.0

    for i in range(1,len(poses)):
        tgt_pose = poses[i]
        tgt_camera_center = null_space(tgt_pose[:3,:])
        tgt_camera_center = tgt_camera_center[:3,0] / tgt_camera_center[3,0]
        b = np.linalg.norm(src_camera_center - tgt_camera_center)
        if b < min_baseline:
            min_baseline = b
        if b > max_baseline:
            max_baseline = b
    return min_baseline, max_baseline

def camera_center(cam: np.ndarray) -> np.ndarray:
    """Computes the center of a camera in world coordinates.

    Args:
        cam: The extrinsics matrix (4x4) of a given camera.

    Returns:
        The camera center vector (3x1) in world coordinates.
    """
    C = null_space(cam[:3,:4])
    C /= C[3,:]

    return C

def relative_transform(cams_1: np.ndarray, cams_2: np.ndarray) -> np.ndarray:
    """Computes the relative transformation between two sets of cameras.

    Args:
        cams_1: Array of the first set of cameras (Nx4x4).
        cams_2: Array of the second set of cameras (Nx4x4).

    Returns:
        The relative transformation matrix (4x4) between the two trajectories.
    """
    centers_1 = np.squeeze(np.array([ camera_center(c) for c in cams_1 ]), axis=2)
    centers_2 = np.squeeze(np.array([ camera_center(c) for c in cams_2 ]), axis=2)

    ### determine scale
    # grab first camera pair
    c1_0 = centers_1[0][:3]
    c2_0 = centers_2[0][:3]

    # grab one-hundreth camera pair
    c1_1 = centers_1[99][:3]
    c2_1 = centers_2[99][:3]

    # calculate the baseline between both sets of cameras
    baseline_1 = np.linalg.norm(c1_0 - c1_1)
    baseline_2 = np.linalg.norm(c2_0 - c2_1)

    # compute the scale based on the baseline ratio
    scale = baseline_2/baseline_1

    ### determine 1->2 Rotation 
    b1 = np.array([c[:3] for c in centers_1])
    b2 = np.array([c[:3] for c in centers_2])
    R = Rotation.align_vectors(b2,b1)[0].as_matrix()
    R = scale * R

    ### create transformation matrix
    M = np.eye(4)
    M[:3,:3] = R

    ### determine 1->2 Translation
    num_cams = len(cams_1)
    t = np.array([ c2-(M@c1) for c1,c2 in zip(centers_1,centers_2) ])
    t = np.mean(t, axis=0)

    ### add translation
    M[:3,3] = t[:3]

    return M


def sfm_to_trajectory(cams: np.ndarray, log_file: str) -> None:
    """Convert a set of cameras from SFM format to Trajectory File format.

    Args:
        cams: Array of camera extrinsics (Nx4x4) to be converted.
        log_file: Output path to the *.log file that is to be created.
    """
    num_cams = len(cams)

    with open(log_file, 'w') as f:
        for i,cam in enumerate(cams):
            # write camera to output_file
            f.write("{} {} 0\n".format(str(i),str(i)))
            for row in cam:
                for c in row:
                    f.write("{} ".format(str(c)))
                f.write("\n")
        
    return

def trajectory_to_sfm(log_file: str, camera_path: str, intrinsics: np.ndarray) -> None:
    """Convert a set of cameras from Trajectory File format to SFM format.

    Args:
        log_file: Input *.log file that stores the trajectory information.
        camera_path: Output path where the SFM camera files will be written.
        intrinsics: Array of intrinsics matrices (Nx3x3) for each camera.
    """
    with open(log_file, 'r') as f:
        lines = f.readlines()
        num_lines = len(lines)
        i = 0

        while(i < num_lines-5):
            view_num = int(lines[i].strip().split(" ")[0])
            
            cam = np.zeros((2,4,4))
            cam[0,0,:] = np.asarray(lines[i+1].strip().split(" "), dtype=float)
            cam[0,1,:] = np.asarray(lines[i+2].strip().split(" "), dtype=float)
            cam[0,2,:] = np.asarray(lines[i+3].strip().split(" "), dtype=float)
            cam[0,3,:] = np.asarray(lines[i+4].strip().split(" "), dtype=float)
            cam[0,:,:] = np.linalg.inv(cam[0,:,:])

            cam_file = "{:08d}_cam.txt".format(view_num)
            cam_path = os.path.join(camera_path, cam_file)

            cam[1,:,:] = intrinsics[view_num]

            write_cam(cam_path, cam)
            i = i+5
    return

def y_axis_rotation(P: np.ndarray, theta: float) -> np.ndarray:
    """Applies a rotation to the given camera extrinsics matrix along the y-axis.

    Parameters:
        P: Initial extrinsics camera matrix.
        theta: Angle (in radians) to rotate the camera.

    Returns:
        The rotated extrinsics matrix for the camera.
    """
    R = np.eye(4)
    R[0,0] = math.cos(theta)
    R[0,2] = math.sin(theta)
    R[2,0] = -(math.sin(theta))
    R[2,2] = math.cos(theta)

    P_rot = R @ P

    return P_rot








# cvt/filtering.py

"""A suite of common filtering utilities.

This module includes several functions for filtering depth maps.

This module contains the following functions:

- `conf_filter(depth_map, conf_map, device, min_conf)` - Filters a map by confidence values above a minimum threshold.
- `geometric_filter(src_depth, src_cam, tgt_depth, tgt_cam, pix_th, depth_th)` - Computes a geometric filter based off of pixel and depth reprojection error.
- `topk_filter(depth_map, conf_map, device, percent)` - Filters a map by the top percentage of confidence values.
- `topk_strict_filter(depth_map, filter_prob, device, percent)` - Filters a map by the top percentage of confidence values.
"""

import numpy as np
import torch
import os
from typing import Tuple
from tqdm import tqdm
import cv2
from PIL import Image

from cvt.geometry import reproject
from cvt.io import read_cluster_list, write_pfm, read_pfm

def conf_filter(depth_map: torch.Tensor, conf_map: torch.Tensor, device: str = 'cuda:0', min_conf: float = 0.8) -> Tuple[torch.Tensor, torch.Tensor]:
    """Filters a map by confidence values above a minimum threshold.

    Parameters:
        depth_map:
        conf_map:
        device:
        min_conf:

    Returns:
        filtered_map:
        mask:
    """
    mask = (torch.ge(conf_map, min_conf)).to(torch.float32).to(device)
    return depth_map*mask, mask

def geometric_filter(src_depth: np.ndarray, src_cam: np.ndarray, tgt_depth: np.ndarray, tgt_cam: np.ndarray, pix_th: float=1.0, depth_th: float=0.01) -> Tuple[np.ndarray,np.ndarray,np.ndarray]:
    """Computes a geometric filter based off of pixel and depth reprojection error.

    Parameters:
        src_depth:
        src_cam:
        tgt_depth:
        tgt_depth:
        pix_th:
        depth_th:

    Returns:
        mask:
        depth_reprojected:
        coords_tgt:
    """
    width, height = src_depth.shape[1], src_depth.shape[0]
    x_ref, y_ref = np.meshgrid(np.arange(0, width), np.arange(0, height))
    depth_reprojected, coords_reprojected, coords_tgt = reproject(src_depth, src_cam, tgt_depth, tgt_cam)
    
    # measure pixel difference
    dist = np.sqrt((coords_reprojected[:,:,0] - x_ref) ** 2 + (coords_reprojected[:,:,1] - y_ref) ** 2)

    # measure depth difference
    depth_diff = np.abs(depth_reprojected - src_depth)
    relative_depth_diff = depth_diff / src_depth

    # compute mask
    mask = np.logical_and(dist < pix_th, relative_depth_diff < depth_th)

    # apply mask to depth map
    depth_reprojected[~mask] = 0

    return mask, depth_reprojected, coords_tgt

def topk_filter(depth_map: np.ndarray, conf_map:np.ndarray, device: str='cuda:0', percent: float=0.3) -> Tuple[np.ndarray,np.ndarray]:
    """Filters a map by the top percentage of confidence values.

    Parameters:
        depth_map:
        conf_map:
        device:
        percent:

    Returns:
        filtered_map:
        mask:
    """
    height, width = depth_map.shape

    # calculate k number of points to keep
    valid_map = torch.ne(conf_map, 0.0).to(torch.float32)
    valid_count = torch.sum(valid_map)
    k = int(percent * valid_count)

    # flatten and grab top-k indices
    filter_prob = conf_map.reshape(-1)
    (vals, indices) = torch.topk(filter_prob, k=k, dim=0)

    # get min confidence value
    min_conf = torch.min(vals)

    # filter by min conf value
    filt = (torch.ge(conf_map, min_conf)).to(torch.float32).to(device)

    return depth_map*filt, filt

def topk_strict_filter(depth_map: np.ndarray, filter_prob: np.ndarray, device: str = 'cuda:0', percent: float = 0.3) -> Tuple[np.ndarray,np.ndarray]:
    """Filters a map by the top percentage of confidence values.

    Parameters:
        depth_map:
        filter_prob:
        device:
        percent:

    Returns:
        filtered_map:
        mask:
    """
    height, width = depth_map.shape

    # calculate k number of points to keep
    valid_map = torch.ne(filter_prob, 0.0).to(torch.float32)
    valid_count = torch.sum(valid_map)
    k = int(percent * valid_count)

    # flatten and grab top-k indices
    filter_prob = filter_prob.reshape(-1)
    (vals, indices) = torch.topk(filter_prob, k=k, dim=0)

    # calculate the row and column given each index
    row_indices = torch.div(indices, width, rounding_mode="floor").unsqueeze(-1)
    col_indices = torch.remainder(indices, width).unsqueeze(-1)

    # concatenate the [r,c] indices into a single tensor
    indices = torch.cat((row_indices, col_indices), dim=1)
    filt = torch.zeros((height,width), dtype=torch.uint8).to(device)

    # set top-k indices to 1
    for r,c in indices:
        filt[r,c] = 1

    return depth_map*filt, filt


# read a binary mask
def read_mask(filename):
    return read_img(filename) > 0.5


# save a binary mask
def save_mask(filename, mask):
    assert mask.dtype == bool
    mask = mask.astype(np.uint8) * 255
    Image.fromarray(mask).save(filename)


# read a pair file, [(ref_view1, [src_view1-1, ...]), (ref_view2, [src_view2-1, ...]), ...]
def read_pair_file(filename):
    data = []
    all_views = list(range(0,49))
    with open(filename) as f:
        num_viewpoint = int(f.readline())
        # 49 viewpoints
        for view_idx in range(num_viewpoint):
            ref_view = int(f.readline().rstrip())
            src_views = [int(x) for x in f.readline().rstrip().split()[1::2]]
            if len(src_views) == 0:
                continue
            data.append((ref_view, src_views))
    return data

# project the reference point cloud into the source view, then project back
def reproject_with_depth(depth_ref, intrinsics_ref, extrinsics_ref, depth_src, intrinsics_src, extrinsics_src):
    width, height = depth_ref.shape[1], depth_ref.shape[0]
    ## step1. project reference pixels to the source view
    # reference view x, y
    x_ref, y_ref = np.meshgrid(np.arange(0, width), np.arange(0, height))
    x_ref, y_ref = x_ref.reshape([-1]), y_ref.reshape([-1])
    # reference 3D space
    xyz_ref = np.matmul(np.linalg.inv(intrinsics_ref),
                        np.vstack((x_ref, y_ref, np.ones_like(x_ref))) * depth_ref.reshape([-1]))
    # source 3D space
    xyz_src = np.matmul(np.matmul(extrinsics_src, np.linalg.inv(extrinsics_ref)),
                        np.vstack((xyz_ref, np.ones_like(x_ref))))[:3]
    # source view x, y
    K_xyz_src = np.matmul(intrinsics_src, xyz_src)
    xy_src = K_xyz_src[:2] / K_xyz_src[2:3]

    ## step2. reproject the source view points with source view depth estimation
    # find the depth estimation of the source view
    x_src = xy_src[0].reshape([height, width]).astype(np.float32)
    y_src = xy_src[1].reshape([height, width]).astype(np.float32)
    sampled_depth_src = cv2.remap(depth_src, x_src, y_src, interpolation=cv2.INTER_LINEAR)
    # mask = sampled_depth_src > 0

    # source 3D space
    # NOTE that we should use sampled source-view depth_here to project back
    xyz_src = np.matmul(np.linalg.inv(intrinsics_src),
                        np.vstack((xy_src, np.ones_like(x_ref))) * sampled_depth_src.reshape([-1]))
    # reference 3D space
    xyz_reprojected = np.matmul(np.matmul(extrinsics_ref, np.linalg.inv(extrinsics_src)),
                                np.vstack((xyz_src, np.ones_like(x_ref))))[:3]
    # source view x, y, depth
    depth_reprojected = xyz_reprojected[2].reshape([height, width]).astype(np.float32)
    K_xyz_reprojected = np.matmul(intrinsics_ref, xyz_reprojected)
    xy_reprojected = K_xyz_reprojected[:2] / (K_xyz_reprojected[2:3] + 1e-7)
    x_reprojected = xy_reprojected[0].reshape([height, width]).astype(np.float32)
    y_reprojected = xy_reprojected[1].reshape([height, width]).astype(np.float32)

    return depth_reprojected, x_reprojected, y_reprojected, x_src, y_src


def check_geometric_consistency(depth_ref, intrinsics_ref, extrinsics_ref, depth_src, intrinsics_src, extrinsics_src, pix_th=1.0):
    width, height = depth_ref.shape[1], depth_ref.shape[0]
    x_ref, y_ref = np.meshgrid(np.arange(0, width), np.arange(0, height))
    depth_reprojected, x2d_reprojected, y2d_reprojected, x2d_src, y2d_src = reproject_with_depth(depth_ref, intrinsics_ref, extrinsics_ref,
                                                     depth_src, intrinsics_src, extrinsics_src)
    # check |p_reproj-p_1| < 1
    dist = np.sqrt((x2d_reprojected - x_ref) ** 2 + (y2d_reprojected - y_ref) ** 2)

    # check |d_reproj-d_1| / d_1 < 0.01
    depth_diff = np.abs(depth_reprojected - depth_ref)
    relative_depth_diff = depth_diff / depth_ref

    mask = dist < pix_th
    depth_reprojected[~mask] = 0

    return mask, depth_reprojected, x2d_src, y2d_src

def consensus_filter(cfg, est_depth_path, est_conf_path, rgb_path, output_path, dataset, scene):
    K = dataset.K[scene]
    poses = dataset.get_all_poses(scene)
    pix_th = cfg["point_cloud"]["pix_th"]
    prob_th = cfg["point_cloud"]["prob_th"]
    num_consistent = cfg["point_cloud"]["num_consistent"]

    vertexs = []
    vertex_colors = []

    clusters = read_cluster_list(dataset.get_cluster_file(scene))
    nviews = len(clusters)

    out_mask_path = os.path.join(output_path, "masks")
    out_geo_mask_path = os.path.join(out_mask_path, "geometric")
    out_conf_mask_path = os.path.join(out_mask_path, "confidence")
    os.makedirs(out_mask_path, exist_ok=True)
    os.makedirs(out_geo_mask_path, exist_ok=True)
    os.makedirs(out_conf_mask_path, exist_ok=True)

    # for each reference view and the corresponding source views
    for (ref_frame, src_frames) in tqdm(clusters, desc="Building point cloud", unit="views"):
        ref_image = cv2.imread(os.path.join(rgb_path, f"{ref_frame:08d}.png"))
        ref_image = cv2.cvtColor(ref_image, cv2.COLOR_BGR2RGB ) 
        ref_depth_est = read_pfm(os.path.join(est_depth_path, f"{ref_frame:08d}.pfm"))
        ref_pose = poses[ref_frame]
        if prob_th == 0.0:
            conf_mask = (np.ones(ref_depth_est.shape, dtype=np.int64) == 1)
        else:
            confidence = read_pfm(os.path.join(est_conf_path, f"{ref_frame:08d}.pfm"))
            conf_mask = confidence > prob_th
            if conf_mask.shape[0] != ref_depth_est.shape[0] or conf_mask.shape[1] != ref_depth_est.shape[1]:
                conf_mask_t = torch.tensor(conf_mask, dtype=torch.float32)
                conf_mask = torch.squeeze(
                    F.interpolate(
                        torch.unsqueeze(torch.unsqueeze(conf_mask_t, 0), 0), 
                        [ref_depth_est.shape[0], ref_depth_est.shape[1]], mode="nearest")).numpy() == 1.0

        all_srcview_depth_ests = []
        all_srcview_x = []
        all_srcview_y = []
        all_srcview_geomask = []

        ### Geometric Mask ###
        geo_mask_sum = 0
        for src_frame in src_frames:
            src_pose = poses[src_frame]
            src_depth_est = read_pfm(os.path.join(est_depth_path, f"{src_frame:08d}.pfm"))

            # compute geometric mask
            geo_mask, depth_reprojected, x2d_src, y2d_src = check_geometric_consistency(ref_depth_est, K, ref_pose, src_depth_est, K, src_pose, pix_th=pix_th)
            geo_mask_sum += geo_mask.astype(np.int32)
            all_srcview_depth_ests.append(depth_reprojected)
            all_srcview_x.append(x2d_src)
            all_srcview_y.append(y2d_src)
            all_srcview_geomask.append(geo_mask)

        depth_est_averaged = (sum(all_srcview_depth_ests) + ref_depth_est) / (geo_mask_sum + 1)
        geo_mask = geo_mask_sum >= num_consistent
        final_mask = np.logical_and(conf_mask, geo_mask)

        save_mask(os.path.join(out_conf_mask_path, f"{ref_frame:08d}_conf.png"), conf_mask)
        save_mask(os.path.join(out_geo_mask_path, f"{ref_frame:08d}_geo.png"), geo_mask)
        save_mask(os.path.join(out_mask_path, f"{ref_frame:08d}.png"), final_mask)

        height, width = depth_est_averaged.shape[:2]
        x, y = np.meshgrid(np.arange(0, width), np.arange(0, height))
        valid_points = final_mask

        # use either average or reference depth estimates
        #x, y, depth = x[valid_points], y[valid_points], depth_est_averaged[valid_points]
        x, y, depth = x[valid_points], y[valid_points], ref_depth_est[valid_points]

        color = ref_image[valid_points]
        xyz_ref = np.matmul(np.linalg.inv(K),
                            np.vstack((x, y, np.ones_like(x))) * depth)
        xyz_world = np.matmul(np.linalg.inv(ref_pose),
                              np.vstack((xyz_ref, np.ones_like(x))))[:3]
        vertexs.append(xyz_world.transpose((1, 0)))
        vertex_colors.append((color[:,:3]).astype(np.uint8))

    vertexs = np.concatenate(vertexs, axis=0)
    vertex_colors = np.concatenate(vertex_colors, axis=0)
    vertexs = np.array([tuple(v) for v in vertexs], dtype=[('x', 'f4'), ('y', 'f4'), ('z', 'f4')])
    vertex_colors = np.array([tuple(v) for v in vertex_colors], dtype=[('red', 'u1'), ('green', 'u1'), ('blue', 'u1')])

    vertex_all = np.empty(len(vertexs), vertexs.dtype.descr + vertex_colors.dtype.descr)
    for prop in vertexs.dtype.names:
        vertex_all[prop] = vertexs[prop]
    for prop in vertex_colors.dtype.names:
        vertex_all[prop] = vertex_colors[prop]

    el = PlyElement.describe(vertex_all, 'vertex')

    # save point cloud to scene-level points path
    ply_file_scene = os.path.join(output_path, f"{scene}.ply")
    PlyData([el]).write(ply_file_scene)


def reprojection_error(cfg, est_depth_path, rgb_path, out_error_path, dataset, scene, max_views=10):
    K = dataset.K[scene]
    poses = dataset.get_all_poses(scene)

    clusters = read_cluster_list(dataset.get_cluster_file(scene))
    nviews = len(clusters)

    # for each reference view and the corresponding source views
    for (ref_frame, src_frames) in tqdm(clusters, desc="Computing reprojection error", unit="views"):
        ref_image = cv2.imread(os.path.join(rgb_path, f"{ref_frame:08d}.png"))
        ref_image = cv2.cvtColor(ref_image, cv2.COLOR_BGR2RGB ) 
        ref_depth_est = read_pfm(os.path.join(est_depth_path, f"{ref_frame:08d}.pfm"))
        height,width = ref_depth_est.shape
        ref_pose = poses[ref_frame]

        ### Geometric Mask ###
        errors = []
        for v,src_frame in enumerate(src_frames):
            src_pose = poses[src_frame]
            src_depth_est = read_pfm(os.path.join(est_depth_path, f"{src_frame:08d}.pfm"))

            # compute geometric mask
            width, height = ref_depth_est.shape[1], ref_depth_est.shape[0]
            x_ref, y_ref = np.meshgrid(np.arange(0, width), np.arange(0, height))
            _, x2d_reprojected, y2d_reprojected, _, _ = reproject_with_depth(
                                                            ref_depth_est,
                                                            K,
                                                            ref_pose,
                                                            src_depth_est,
                                                            K,
                                                            src_pose)

            errors.append((np.sqrt((x2d_reprojected - x_ref) ** 2 + (y2d_reprojected - y_ref) ** 2)).astype(np.float32))


        errors = np.asarray(errors)
        errors = np.sort(errors,axis=0)
        np.save(os.path.join(out_error_path, f"{ref_frame:08d}.npy"), errors)

        errors = errors[:max_views]

        #for v in range(max_views):
        #    write_pfm(os.path.join(out_error_path, f"{ref_frame:08d}_{v}.pfm"), errors[v])
        #    # visualize clipped version
        #    e = np.clip(errors[v], 0, 2)
        #    cv2.imwrite(os.path.join(out_error_path, f"{ref_frame:08d}_{v}.png"), (e-e.min())/(e.max()-e.min())*255)

        error_sum = errors.sum(axis=0)
        write_pfm(os.path.join(out_error_path, f"{ref_frame:08d}_sum.pfm"), error_sum)

        # visualize clipped version
        e = np.clip(error_sum, 0, 20)
        cv2.imwrite(os.path.join(out_error_path, f"{ref_frame:08d}_sum.png"), (e-e.min())/(e.max()-e.min())*255)

def laplacians(cfg, paths):

    # get depth map filenames
    est_depth_files = os.listdir(paths["depth"])
    est_depth_files = [edf for edf in est_depth_files if edf[-3:] == "pfm" ]
    est_depth_files.sort()

    for i, edf in enumerate(est_depth_files):
        ref_ind = int(edf[:8])
        image = cv2.imread(os.path.join(paths["rgb"], f"{ref_ind:08d}.png"))
        image = torch.movedim(torch.tensor(cv2.cvtColor(image, cv2.COLOR_BGR2RGB)), (0,1,2),(1,2,0))
        c,height,width = image.shape

        image = image.reshape(1,c,height,width).to(torch.float32)
        est_depth = torch.tensor(read_pfm(os.path.join(paths["depth"], edf))).reshape(1,1,height,width).to(torch.float32)

        image_laplacian = laplacian_pyramid(image)
        est_depth_laplacian = laplacian_pyramid(est_depth)

        img_lap = image_laplacian[0,0].detach().cpu().numpy()
        img_lap_filename = os.path.join(paths["laplacian"], f"{ref_ind:08d}_image.pfm")
        write_pfm(img_lap_filename, img_lap)
        depth_lap = est_depth_laplacian[0,0].detach().cpu().numpy()
        depth_lap_filename = os.path.join(paths["laplacian"], f"{ref_ind:08d}_depth.pfm")
        write_pfm(depth_lap_filename, depth_lap)



import os
import random
import numpy as np
import torch
import cv2
import sys
import json
from tqdm import tqdm
import yaml
import scipy.io as sio

from cvt.camera import scale_cam
from cvt.io import read_single_cam_sfm, read_pfm, read_cluster_list
from cvt.datasets.BaseDataset import BaseDataset
from cvt.common import round_nearest

class Dataset(BaseDataset):
    def __init__(self, cfg, mode, scenes):
        super(Dataset, self).__init__(cfg, mode, scenes)
        self.units = "mm"
        self.cameras_path = os.path.join(self.data_path, "Cameras")
        self.depths_path = os.path.join(self.data_path, "Depths")
        self.images_path = os.path.join(self.data_path, "Images")
        self.meshes_path = os.path.join(self.data_path, "Meshes")
        self.points_path = os.path.join(self.data_path, "Points")

    def get_frame_count(self, scene):
        image_files = os.listdir(os.path.join(self.data_path, "Images", scene))
        image_files = [img for img in image_files if img[-4:]==".png"]
        return len(image_files)

    def build_samples(self):
        self.total_samples = []
        self.frame_count = 0

        if self.mode=="inference":
            scenes = self.scenes
        else:
            scenes = tqdm(self.scenes, desc="Loading scene paths", unit="scene")

        for scene in scenes:
            if not os.path.isdir(os.path.join(self.data_path, "Images", scene)):
                print(f"{os.path.join(self.data_path, scene)} is not a valid directory.")
                continue

            # build samples dict for other data
            curr_frame_count = self.get_frame_count(scene)
            if self.sample_mode=="stream":
                self.total_samples.extend(self.build_samples_stream(scene, curr_frame_count))
            elif self.sample_mode=="cluster":
                self.total_samples.extend(self.build_samples_cluster(scene))
            self.frame_count += curr_frame_count

            # load pose and intrinsics for current scene
            self.load_intrinsics(scene)

        self.total_samples = np.asarray(self.total_samples)
        return


    def build_samples_stream(self, scene, frame_count):
        samples = []

        frame_offset = ((self.num_frame-1)//2)
        radius = frame_offset*self.frame_spacing
        for ref_frame in range(frame_count):
            skip = False
            if ((ref_frame + radius >= frame_count) or (ref_frame - radius < 0)):
                continue
            start = ref_frame - radius
            end = ref_frame + radius

            frame_inds = [i for i in range(ref_frame, end+1, self.frame_spacing) if i != ref_frame]
            bottom = [i for i in range(ref_frame, start-1, -self.frame_spacing) if i != ref_frame]
            frame_inds.extend(bottom)
            frame_inds.insert(0, ref_frame)

            image_files = []
            depth_files = []
            pose_files = []
            for ind in frame_inds:
                image_files.append(os.path.join(self.data_path, "Images", scene, f"{ind:08d}.png"))
                depth_files.append(os.path.join(self.data_path, "GT_Depths", scene, f"{ind:08d}_depth.pfm"))
                pose_files.append(os.path.join(self.data_path, "Cameras", f"{ind:08d}_cam.txt"))
                if not os.path.isfile(pose_files[-1]):
                    #print(f"{pose_files[-1]} does not exist; skipping")
                    skip = True
            if skip:
                continue

            samples.append({"scene": scene,
                            "frame_inds": frame_inds,
                            "image_files": image_files,
                            "depth_files": depth_files,
                            "pose_files": pose_files
                            })
        return samples

    def get_cluster_file(self, scene):
        return os.path.join(self.data_path, "Cameras/pair.txt")

    def build_samples_cluster(self, scene):
        samples = []
        clusters = read_cluster_list(self.get_cluster_file(scene))

        for c in clusters:
            skip = False
            frame_inds = [c[0]]
            if self.random_clusters:
                src_frames = np.random.permutation(c[1])
            else:
                src_frames = c[1]
            frame_inds.extend(src_frames[:self.num_frame-1])

            image_files = []
            depth_files = []
            pose_files = []
            for ind in frame_inds:
                image_files.append(os.path.join(self.data_path, "Images", scene, f"{ind:08d}.png"))
                depth_files.append(os.path.join(self.data_path, "GT_Depths", scene, f"{ind:08d}_depth.pfm"))
                pose_files.append(os.path.join(self.data_path, "Cameras", f"{ind:08d}_cam.txt"))
                if not os.path.isfile(pose_files[-1]):
                    #print(f"{pose_files[-1]} does not exist; skipping")
                    skip = True
            if skip:
                continue

            samples.append({"scene": scene,
                            "frame_inds": frame_inds,
                            "image_files": image_files,
                            "depth_files": depth_files,
                            "pose_files": pose_files
                            })
        return samples


    def load_intrinsics(self, scene):
        cam_file = os.path.join(self.data_path, "Cameras/00000000_cam.txt")
        cam = read_single_cam_sfm(cam_file)
        self.K[scene] = cam[1,:3,:3]
        self.K[scene][0,2] -= (self.crop_w//2)
        self.K[scene][1,2] -= (self.crop_h//2)
        self.H = int(self.scale * (self.cfg["camera"]["height"] - self.crop_h))
        self.W = int(self.scale * (self.cfg["camera"]["width"]- self.crop_w))

    def get_pose(self, pose_file, frame_id=None):
        try:
            cam = read_single_cam_sfm(pose_file)
        except:
            print(f"Pose file {pose_file} does not exist.")
            sys.exit()
        pose = cam[0]
        if(np.isnan(pose).any()):
            print(pose, pose_file)
        return pose

    def get_all_poses(self, scene):
        poses = []
        pose_path = os.path.join(self.data_path, "Cameras")
        pose_files = os.listdir(pose_path)
        pose_files.sort()
        for pose_file in pose_files:
            if (pose_file[-8:] != "_cam.txt"):
                continue
            cam = read_single_cam_sfm(os.path.join(pose_path,pose_file))
            pose = cam[0]
            if(np.isnan(pose).any()):
                print(pose, pose_file)
            poses.append(pose)
        return poses

    def get_all_depths(self, scale):
        gt_depth_files = os.listdir(self.gt_depth_path)
        gt_depth_files = [os.path.join(self.gt_depth_path, gdf) for gdf in gt_depth_files if os.path.isfile(os.path.join(self.gt_depth_path, gdf))]
        gt_depth_files.sort()

        gt_depths = {}
        for gdf in gt_depth_files:
            ref_ind = int(gdf[-18:-10])
            gt_depth = self.get_depth(os.path.join(self.gt_depth_path, gdf), scale=scale)
            gt_depths[ref_ind] = gt_depth
        return gt_depths

def build_est_points_filter(points, mask_path, scan_num):
    # read in matlab bounding box, mask, and resolution
    mask_filename = "ObsMask{}_10.mat".format(scan_num)
    mask_path = os.path.join(mask_path, mask_filename)
    data = sio.loadmat(mask_path)
    bounds = np.asarray(data["BB"])
    min_bound = bounds[0,:]
    max_bound = bounds[1,:]
    mask = np.asarray(data["ObsMask"])
    res = int(data["Res"])

    points = points.transpose()
    shape = points.shape
    mask_shape = mask.shape
    filt = np.zeros(shape[1])

    min_bound = min_bound.reshape(3,1)
    min_bound = np.tile(min_bound, (1,shape[1]))

    qv = points
    qv = (points - min_bound) / res
    qv = round_nearest(qv).astype(int)

    # get all valid points
    in_bounds = np.asarray(np.where( ((qv[0,:]>=0) & (qv[0,:] < mask_shape[0]) & (qv[1,:]>=0) & (qv[1,:] < mask_shape[1]) & (qv[2,:]>=0) & (qv[2,:] < mask_shape[2])))).squeeze(0)
    valid_points = qv[:,in_bounds]

    # convert 3D coords ([x,y,z]) to appropriate flattened coordinate ((x*mask_shape[1]*mask_shape[2]) + (y*mask_shape[2]) + z )
    mask_inds = np.ravel_multi_index(valid_points, dims=mask.shape, order='C')

    # further trim down valid points by mask value (keep point if mask is True)
    mask = mask.flatten()
    valid_mask_points = np.asarray(np.where(mask[mask_inds] == True)).squeeze(0)

    # add 1 to indices where we want to keep points
    filt[in_bounds[valid_mask_points]] = 1

    return filt

def build_gt_points_filter(cloud, data_path, scan_num):
    # read in matlab gt plane 
    mask_filename = "Plane{}.mat".format(scan_num)
    mask_path = os.path.join(data_path, "ObsMask", mask_filename)
    data = sio.loadmat(mask_path)
    P = np.asarray(data["P"])

    points = np.asarray(cloud.points).transpose()
    shape = points.shape

    # compute iner-product between points and the defined plane
    Pt = P.transpose()

    points = np.concatenate((points, np.ones((1,shape[1]))), axis=0)
    plane_prod = (Pt @ points).squeeze(0)

    # get all valid points
    filt = np.asarray(np.where((plane_prod > 0), 1, 0))

    return filt



import os
import random
import numpy as np
import torch
import torch.utils.data as data
import cv2
import sys

from cvt.io import read_pfm
from cvt.camera import compute_baselines, _intrinsic_pyramid, scale_cam, crop_cam
from cvt.common import _build_depth_pyramid, _normalize_image, crop_image

def build_dataset(cfg, mode, scenes):
    if cfg["dataset"] == 'DTU':
        from cvt.datasets.DTU import Dataset
    else:
        raise Exception(f"Unknown Dataset {self.cfg['dataset']}")

    return Dataset(cfg, mode, scenes)

class BaseDataset(data.Dataset):
    def __init__(self, cfg, mode, scenes):
        self.cfg = cfg
        self.mode = mode
        self.data_path = self.cfg["data_path"]
        self.device = self.cfg["device"]
        self.scenes = scenes
        self.crop_h = self.cfg["camera"]["crop_h"]
        self.crop_w = self.cfg["camera"]["crop_w"]

        try:
            self.resolution_levels = len(self.cfg["model"]["feature_channels"])
        except:
            self.resolution_levels = 1

        if self.mode=="inference":
            self.num_frame = self.cfg["inference"]["num_frame"]
            self.frame_spacing = self.cfg["inference"]["frame_spacing"]
            self.scale = self.cfg["inference"]["scale"]
            self.sample_mode = self.cfg["inference"]["sample_mode"]
            self.random_crop = False
        else:
            self.num_frame = self.cfg["training"]["num_frame"]
            self.frame_spacing = self.cfg["training"]["frame_spacing"]
            self.scale = self.cfg["training"]["scale"]
            self.sample_mode = self.cfg["training"]["sample_mode"]
            self.random_crop = self.cfg["training"]["random_crop"]
        self.random_clusters = (mode=="training")

        self.K = {}
        self.build_samples()

        if self.mode == "inference":
            # use all samples during inference
            self.samples = self.total_samples
        else:
            # shuffle and sub-sample during training
            if self.mode=="training":
                self.max_samples = self.cfg["training"]["max_training_samples"]
            elif self.mode=="validation":
                self.max_samples = self.cfg["training"]["max_val_samples"]
            self.shuffle_and_subsample()

    def build_samples(self, frame_spacing):
        raise NotImplementedError()

    def load_intrinsics(self):
        raise NotImplementedError()

    def get_pose(self, frame_id):
        raise NotImplementedError()

    def get_image(self, image_file, scale=True):
        image = cv2.imread(image_file)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # crop and resize image
        h,w,_ = image.shape
        image = image[(self.crop_h//2):h-(self.crop_h//2),(self.crop_w//2):w-(self.crop_w//2), :]
        if scale:
            image = cv2.resize(image, (self.W,self.H), interpolation=cv2.INTER_LINEAR)
        image = _normalize_image(image, mean=0.5, std=0.5)
        image = np.moveaxis(image, [0,1,2], [1,2,0])
        return image.astype(np.float32)

    def get_depth(self, depth_file, scale=True):
        if (depth_file[-3:] == "pfm"):
            depth = read_pfm(depth_file)
        elif (depth_file[-3:] == "png"):
            depth = cv2.imread(depth_file, 2) / self.png_depth_scale

        # crop and resize depth
        h,w = depth.shape
        depth = depth[(self.crop_h//2):h-(self.crop_h//2),(self.crop_w//2):w-(self.crop_w//2)]
        if scale:
            depth = cv2.resize(depth, (self.W,self.H), interpolation=cv2.INTER_LINEAR)
        depth = depth.reshape(1, depth.shape[0], depth.shape[1])
        return depth.astype(np.float32)

    def shuffle_and_subsample(self):
        np.random.shuffle(self.total_samples)
        num_samples = min(len(self.total_samples), self.max_samples)
        self.samples = self.total_samples[:num_samples]

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        idx = idx % self.__len__()
        sample = self.samples[idx]
        scene = sample["scene"]

        # load and compute intrinsics
        K = np.copy(self.K[scene])

        # random crop scaled image patches or scale entire image
        if self.random_crop:
            crop_row = np.random.randint(0, (self.cfg["camera"]["height"] - self.crop_h) - self.H)
            crop_col = np.random.randint(0, (self.cfg["camera"]["width"]- self.crop_w) - self.W)
            K = crop_cam(K, crop_row, crop_col)
        else:
            K = scale_cam(K, scale=self.scale)

        images = [None]*self.num_frame
        poses = [None]*self.num_frame
        target_depths = [None]*self.num_frame
        filenames = []
        for i, fid in enumerate(sample["frame_inds"]):
            images[i] = self.get_image(sample["image_files"][i], scale=(not self.random_crop))
            poses[i] = self.get_pose(sample["pose_files"][i], fid)
            target_depths[i] = self.get_depth(sample["depth_files"][i], scale=(not self.random_crop))
            filenames.append(scene + '-' + '_'.join('%04d' % x for x in sample["frame_inds"]))

            if self.random_crop:
                images[i] = crop_image(images[i], crop_row, crop_col, self.scale)
                target_depths[i] = crop_image(target_depths[i], crop_row, crop_col, self.scale)
        images = np.asarray(images, dtype=np.float32)
        poses = np.asarray(poses, dtype=np.float32)
        target_depths = np.asarray(target_depths, dtype=np.float32)

        # compute min and max camera baselines
        min_baseline, max_baseline = compute_baselines(poses)

        # load data dict
        data = {}
        data["ref_id"] = int(sample["frame_inds"][0])
        data["K"] = K
        data["images"] = images
        data["poses"] = poses
        data["target_depth"] = target_depths[0]
        if self.cfg["camera"]["baseline_mode"] == "min":
            data["baseline"] = min_baseline
        elif self.cfg["camera"]["baseline_mode"] == "max":
            data["baseline"] = max_baseline

        ## Scaling intrinsics for the feature pyramid
        if self.resolution_levels > 1:
            data["target_depths"] = _build_depth_pyramid(data["target_depth"][0], levels=self.resolution_levels)

            multires_intrinsics = []
            for i in range(self.num_frame):
                multires_intrinsics.append(_intrinsic_pyramid(K, self.resolution_levels)[::-1])
            
            data["multires_intrinsics"] = np.stack(multires_intrinsics).astype(np.float32)

        return data



from setuptools import setup, find_packages

with open("README.md", "r") as doc:
    long_desc = doc.read()

setup(
        name="cvtkit",
        description="A Python library including general functions and operations on various computer vision related structures.",
        long_description=long_desc,
        long_description_content_type="text/markdown",
        packages=find_packages(where="src"),
        package_dir={"": "src"},
        install_repuires=[],
        extras_require={
            "dev": ["pytest>=7.0"],
            },
        python_requires=">=3.8"
)



# datasets/__init__.py

"""A collection of dataset-specific functions.

This package includes the following modules:

- `dtu`: A collection of functions for the DTU dataset.
"""
import os
import sys

FILE_ROOT=os.path.abspath(os.path.dirname(__file__))
PYTHON_ROOT=os.path.abspath(os.path.dirname(FILE_ROOT))

sys.path.append(PYTHON_ROOT)

from .DTU import *



# cvt/visualization/latex.py
"""Module including routines for creating latex visuals.

This module includes the following functions:

- `create_subfigures(images, captions, labels)` - Creates a multi-image figure in latex.
"""
import sys
import os
import numpy as np

from typing import List


def create_subfigures(images: List[np.ndarray], captions: List[str], labels: List[str]) -> str:
    """Creates a multi-image figure in latex.

    Parameters:
        images: List of images to include in the figure.
        captions: Corresponding captions for each subfigure.
        labels: Corresponding labels for each subfigure.

    Returns:
        The latex string for the figure.
    """
    num_sub_figs = len(images)
    width_scale = 1/num_sub_figs

    figure = "\\begin{figure}\n"

    for n in range(num_sub_figs):
        figure +=   "\t\\centering\n" + \
                    "\t\\begin{{subfigure}}{{{w:.1f}\\textwidth}}\n".format(w=width_scale) + \
                        "\t\t\\centering\n" + \
                        "\t\t\\includegraphics[width=\\textwidth]{{{img}}}\n".format(img=images[n]) + \
                        "\t\t\\caption{{{caption}}}\n".format(caption=captions[n]) + \
                        "\t\t\\label{{fig:{label}}}\n".format(label=labels[n]) + \
                        "\t\\end{subfigure}\n\t\\hfill\n"
    figure +=   "\t\\caption{{{caption}}}\n".format(caption=captions[-1]) + \
                "\t\\label{{fig:{label}}}\n".format(label=labels[-1]) + \
                "\\end{figure}"

    return figure
